@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2021 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2022 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -56,7 +56,6 @@ u_int8_t		nlisteners;
 int			kore_argc = 0;
 pid_t			kore_pid = -1;
 u_int16_t		cpu_count = 1;
-int			kore_debug = 0;
 int			kore_quiet = 0;
 int			skip_runas = 0;
 int			skip_chroot = 0;
@@ -64,11 +63,11 @@ u_int8_t		worker_count = 0;
 char			**kore_argv = NULL;
 int			kore_foreground = 0;
 char			*kore_progname = NULL;
-char			*kore_root_path = NULL;
-char			*kore_runas_user = NULL;
 u_int32_t		kore_socket_backlog = 5000;
+int			kore_quit = KORE_QUIT_NONE;
 char			*kore_pidfile = KORE_PIDFILE_DEFAULT;
-char			*kore_tls_cipher_list = KORE_DEFAULT_CIPHER_LIST;
+
+struct kore_privsep	worker_privsep;
 
 extern char		**environ;
 extern char		*__progname;
@@ -79,7 +78,7 @@ static void	version(void);
 
 static void	kore_write_kore_pid(void);
 static void	kore_proctitle_setup(void);
-static void	kore_server_sslstart(void);
+static void	kore_server_shutdown(void);
 static void	kore_server_start(int, char *[]);
 static void	kore_call_parent_configure(int, char **);
 
@@ -113,9 +112,9 @@ usage(void)
 #endif
 	printf("\t-f\tstart in foreground\n");
 	printf("\t-h\tthis help text\n");
-	printf("\t-n\tdo not chroot\n");
+	printf("\t-n\tdo not chroot on any worker\n");
 	printf("\t-q\tonly log errors\n");
-	printf("\t-r\tdo not drop privileges\n");
+	printf("\t-r\tdo not change user on any worker\n");
 	printf("\t-v\tdisplay %s build information\n", __progname);
 
 	printf("\nFind more information on https://kore.io\n");
@@ -148,6 +147,8 @@ version(void)
 #if defined(KORE_USE_ACME)
 	printf("acme ");
 #endif
+	if (!kore_tls_supported())
+		printf("notls ");
 	printf("\n");
 	exit(0);
 }
@@ -168,6 +169,9 @@ main(int argc, char *argv[])
 #endif
 
 	kore_mem_init();
+	kore_msg_init();
+	kore_log_init();
+
 	kore_progname = kore_strdup(argv[0]);
 	kore_proctitle_setup();
 
@@ -194,10 +198,6 @@ main(int argc, char *argv[])
 		if (!S_ISDIR(st.st_mode) && !S_ISREG(st.st_mode))
 			fatal("%s: not a directory or file", kore_pymodule);
 	}
-
-#elif !defined(KORE_SINGLE_BINARY)
-	if (argc > 0)
-		fatal("did you mean to run `kodev' instead?");
 #endif
 
 	kore_pid = getpid();
@@ -205,8 +205,6 @@ main(int argc, char *argv[])
 	LIST_INIT(&kore_servers);
 
 	kore_platform_init();
-	kore_log_init();
-	kore_msg_init();
 #if !defined(KORE_NO_HTTP)
 	http_parent_init();
 #if defined(KORE_USE_CURL)
@@ -224,7 +222,7 @@ main(int argc, char *argv[])
 #endif
 	kore_domain_init();
 	kore_module_init();
-	kore_server_sslstart();
+	kore_tls_init();
 
 #if !defined(KORE_SINGLE_BINARY) && !defined(KORE_USE_PYTHON)
 	if (config_file == NULL)
@@ -274,11 +272,7 @@ main(int argc, char *argv[])
 
 	kore_signal_setup();
 	kore_server_start(argc, argv);
-
-	if (!kore_quiet)
-		kore_log(LOG_NOTICE, "server shutting down");
-
-	kore_worker_shutdown();
+	kore_server_shutdown();
 
 	rcall = kore_runtime_getcall(parent_teardown_hook);
 	if (rcall != NULL) {
@@ -292,7 +286,7 @@ main(int argc, char *argv[])
 	kore_server_cleanup();
 
 	if (!kore_quiet)
-		kore_log(LOG_NOTICE, "goodbye");
+		kore_log(LOG_INFO, "goodbye");
 
 #if defined(KORE_USE_PYTHON)
 	kore_python_cleanup();
@@ -300,7 +294,7 @@ main(int argc, char *argv[])
 
 	kore_mem_cleanup();
 
-	return (0);
+	return (kore_quit);
 }
 
 void
@@ -309,9 +303,9 @@ kore_default_getopt(int argc, char **argv)
 	int		ch;
 
 #if !defined(KORE_SINGLE_BINARY)
-	while ((ch = getopt(argc, argv, "c:dfhnqrv")) != -1) {
+	while ((ch = getopt(argc, argv, "c:fhnqrv")) != -1) {
 #else
-	while ((ch = getopt(argc, argv, "dfhnqrv")) != -1) {
+	while ((ch = getopt(argc, argv, "fhnqrv")) != -1) {
 #endif
 		switch (ch) {
 #if !defined(KORE_SINGLE_BINARY)
@@ -320,11 +314,6 @@ kore_default_getopt(int argc, char **argv)
 			if ((config_file = strdup(optarg)) == NULL)
 				fatal("strdup");
 			break;
-#endif
-#if defined(KORE_DEBUG)
-		case 'd':
-			kore_debug = 1;
-			break;
 #endif
 		case 'f':
 			kore_foreground = 1;
@@ -350,74 +339,6 @@ kore_default_getopt(int argc, char **argv)
 	}
 }
 
-int
-kore_tls_sni_cb(SSL *ssl, int *ad, void *arg)
-{
-	struct connection	*c;
-	struct kore_domain	*dom;
-	const char		*sname;
-
-	if ((c = SSL_get_ex_data(ssl, 0)) == NULL)
-		fatal("no connection data in %s", __func__);
-
-	sname = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
-	kore_debug("kore_tls_sni_cb(): received host %s", sname);
-
-	if (sname != NULL)
-		c->tls_sni = kore_strdup(sname);
-
-	if (sname != NULL &&
-	    (dom = kore_domain_lookup(c->owner->server, sname)) != NULL) {
-		if (dom->ssl_ctx == NULL) {
-			kore_log(LOG_NOTICE,
-			    "TLS configuration for %s not complete",
-			    dom->domain);
-			return (SSL_TLSEXT_ERR_NOACK);
-		}
-
-		kore_debug("kore_ssl_sni_cb(): Using %s CTX", sname);
-		SSL_set_SSL_CTX(ssl, dom->ssl_ctx);
-
-		if (dom->cafile != NULL) {
-			SSL_set_verify(ssl, SSL_VERIFY_PEER |
-			    SSL_VERIFY_FAIL_IF_NO_PEER_CERT, NULL);
-			c->flags |= CONN_LOG_TLS_FAILURE;
-		} else {
-			SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);
-		}
-
-#if defined(KORE_USE_ACME)
-		/*
-		 * If ALPN callback was called before SNI was parsed we
-		 * must make sure we swap to the correct certificate now.
-		 */
-		if (c->flags & CONN_TLS_ALPN_ACME_SEEN)
-			kore_acme_tls_challenge_use_cert(ssl, dom);
-
-		c->flags |= CONN_TLS_SNI_SEEN;
-#endif
-		return (SSL_TLSEXT_ERR_OK);
-	}
-
-	return (SSL_TLSEXT_ERR_NOACK);
-}
-
-void
-kore_tls_info_callback(const SSL *ssl, int flags, int ret)
-{
-	struct connection	*c;
-
-	if (flags & SSL_CB_HANDSHAKE_START) {
-		if ((c = SSL_get_app_data(ssl)) == NULL)
-			fatal("no SSL_get_app_data");
-
-#if defined(TLS1_3_VERSION)
-		if (SSL_version(ssl) != TLS1_3_VERSION)
-#endif
-			c->tls_reneg++;
-	}
-}
-
 int
 kore_server_bind(struct kore_server *srv, const char *ip, const char *port,
     const char *ccb)
@@ -426,8 +347,6 @@ kore_server_bind(struct kore_server *srv, const char *ip, const char *port,
 	struct listener		*l;
 	struct addrinfo		hints, *results;
 
-	kore_debug("kore_server_bind(%s, %s, %s)", srv->name, ip, port);
-
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
@@ -497,6 +416,15 @@ kore_server_bind_unix(struct kore_server *srv, const char *path,
 	if (!kore_listener_init(l, AF_UNIX, ccb))
 		return (KORE_RESULT_ERROR);
 
+	if (sun.sun_path[0] != '\0') {
+		if (unlink(sun.sun_path) == -1 && errno != ENOENT) {
+			kore_log(LOG_ERR, "unlink: %s: %s",
+			    sun.sun_path, errno_s);
+			kore_listener_free(l);
+			return (KORE_RESULT_ERROR);
+		}
+	}
+
 	if (bind(l->fd, (struct sockaddr *)&sun, socklen) == -1) {
 		kore_log(LOG_ERR, "bind: %s", errno_s);
 		kore_listener_free(l);
@@ -519,7 +447,11 @@ kore_server_create(const char *name)
 
 	srv = kore_calloc(1, sizeof(struct kore_server));
 	srv->name = kore_strdup(name);
-	srv->tls = 1;
+
+	if (kore_tls_supported())
+		srv->tls = 1;
+	else
+		srv->tls = 0;
 
 	TAILQ_INIT(&srv->domains);
 	LIST_INIT(&srv->listeners);
@@ -545,10 +477,10 @@ kore_server_finalize(struct kore_server *srv)
 			proto = "http";
 
 		if (l->family == AF_UNIX) {
-			kore_log(LOG_NOTICE, "%s serving %s on %s",
+			kore_log(LOG_INFO, "%s serving %s on %s",
 			    srv->name, proto, l->host);
 		} else {
-			kore_log(LOG_NOTICE, "%s serving %s on %s:%s",
+			kore_log(LOG_INFO, "%s serving %s on %s:%s",
 			    srv->name, proto, l->host, l->port);
 		}
 	}
@@ -657,11 +589,30 @@ kore_server_free(struct kore_server *srv)
 void
 kore_listener_free(struct listener *l)
 {
+	int	rm;
+
 	LIST_REMOVE(l, list);
 
 	if (l->fd != -1)
 		close(l->fd);
 
+	rm = 0;
+
+#if defined(__linux__)
+	if (worker == NULL && l->family == AF_UNIX && l->host[0] != '@')
+		rm++;
+#else
+	if (worker == NULL && l->family == AF_UNIX)
+		rm++;
+#endif
+	if (rm) {
+		if (unlink(l->host) == -1) {
+			kore_log(LOG_NOTICE,
+			    "failed to remove unix socket %s (%s)", l->host,
+			    errno_s);
+		}
+	}
+
 	kore_free(l->host);
 	kore_free(l->port);
 
@@ -717,6 +668,23 @@ kore_sockopt(int fd, int what, int opt)
 
 void
 kore_signal_setup(void)
+{
+	kore_signal_trap(SIGHUP);
+	kore_signal_trap(SIGQUIT);
+	kore_signal_trap(SIGTERM);
+	kore_signal_trap(SIGUSR1);
+	kore_signal_trap(SIGCHLD);
+
+	if (kore_foreground)
+		kore_signal_trap(SIGINT);
+	else
+		(void)signal(SIGINT, SIG_IGN);
+
+	(void)signal(SIGPIPE, SIG_IGN);
+}
+
+void
+kore_signal_trap(int sig)
 {
 	struct sigaction	sa;
 
@@ -727,25 +695,8 @@ kore_signal_setup(void)
 	if (sigfillset(&sa.sa_mask) == -1)
 		fatal("sigfillset: %s", errno_s);
 
-	if (sigaction(SIGHUP, &sa, NULL) == -1)
+	if (sigaction(sig, &sa, NULL) == -1)
 		fatal("sigaction: %s", errno_s);
-	if (sigaction(SIGQUIT, &sa, NULL) == -1)
-		fatal("sigaction: %s", errno_s);
-	if (sigaction(SIGTERM, &sa, NULL) == -1)
-		fatal("sigaction: %s", errno_s);
-	if (sigaction(SIGUSR1, &sa, NULL) == -1)
-		fatal("sigaction: %s", errno_s);
-	if (sigaction(SIGCHLD, &sa, NULL) == -1)
-		fatal("sigaction: %s", errno_s);
-
-	if (kore_foreground) {
-		if (sigaction(SIGINT, &sa, NULL) == -1)
-			fatal("sigaction: %s", errno_s);
-	} else {
-		(void)signal(SIGINT, SIG_IGN);
-	}
-
-	(void)signal(SIGPIPE, SIG_IGN);
 }
 
 void
@@ -820,24 +771,49 @@ kore_proctitle_setup(void)
 		proctitle_maxlen += strlen(kore_argv[i]) + 1;
 }
 
-static void
-kore_server_sslstart(void)
-{
-	SSL_library_init();
-	SSL_load_error_strings();
-}
-
 static void
 kore_server_start(int argc, char *argv[])
 {
 	u_int32_t			tmp;
 	struct kore_server		*srv;
 	u_int64_t			netwait;
-	int				quit, last_sig;
+	int				last_sig;
+#if !defined(KORE_NO_HTTP)
+	int				alog;
+	struct kore_domain		*dom;
+#endif
 #if defined(KORE_SINGLE_BINARY)
 	struct kore_runtime_call	*rcall;
 #endif
 
+	if (!kore_quiet) {
+		kore_log(LOG_INFO, "%s %s starting, built=%s",
+		    __progname, kore_version, kore_build_date);
+		kore_log(LOG_INFO, "built-ins: "
+#if defined(__linux__)
+		    "seccomp "
+#endif
+#if defined(KORE_USE_PGSQL)
+		    "pgsql "
+#endif
+#if defined(KORE_USE_TASKS)
+		    "tasks "
+#endif
+#if defined(KORE_USE_JSONRPC)
+		    "jsonrpc "
+#endif
+#if defined(KORE_USE_PYTHON)
+		    "python "
+#endif
+#if defined(KORE_USE_ACME)
+		    "acme "
+#endif
+#if defined(KORE_USE_CURL)
+		    "curl "
+#endif
+		);
+	}
+
 	if (kore_foreground == 0) {
 		if (daemon(1, 0) == -1)
 			fatal("cannot daemon(): %s", errno_s);
@@ -853,25 +829,6 @@ kore_server_start(int argc, char *argv[])
 	kore_pid = getpid();
 	kore_write_kore_pid();
 
-	if (!kore_quiet) {
-		kore_log(LOG_NOTICE, "%s is starting up", __progname);
-#if defined(__linux__)
-		kore_log(LOG_NOTICE, "seccomp sandbox enabled");
-#endif
-#if defined(KORE_USE_PGSQL)
-		kore_log(LOG_NOTICE, "pgsql built-in enabled");
-#endif
-#if defined(KORE_USE_TASKS)
-		kore_log(LOG_NOTICE, "tasks built-in enabled");
-#endif
-#if defined(KORE_USE_JSONRPC)
-		kore_log(LOG_NOTICE, "jsonrpc built-in enabled");
-#endif
-#if defined(KORE_USE_PYTHON)
-		kore_log(LOG_NOTICE, "python built-in enabled");
-#endif
-	}
-
 #if !defined(KORE_SINGLE_BINARY) && !defined(KORE_USE_PYTHON)
 	kore_call_parent_configure(argc, argv);
 #endif
@@ -881,16 +838,36 @@ kore_server_start(int argc, char *argv[])
 		kore_call_parent_configure(argc, argv);
 #endif
 
+#if defined(KORE_USE_PYTHON)
+	kore_python_routes_resolve();
+#endif
+
 	/* Check if keymgr will be active. */
-	LIST_FOREACH(srv, &kore_servers, list) {
-		if (srv->tls) {
-			keymgr_active = 1;
-			break;
+	if (kore_tls_supported()) {
+		LIST_FOREACH(srv, &kore_servers, list) {
+			if (srv->tls) {
+				kore_keymgr_active = 1;
+				break;
+			}
 		}
+	} else {
+		kore_keymgr_active = 0;
 	}
 
 	kore_platform_proctitle("[parent]");
-	kore_worker_init();
+
+	if (!kore_worker_init()) {
+		kore_log(LOG_ERR, "last worker log lines:");
+		kore_log(LOG_ERR, "=====================================");
+		net_init();
+		kore_connection_init();
+		kore_platform_event_init();
+		kore_msg_parent_init();
+		kore_platform_event_wait(10);
+		kore_worker_dispatch_signal(SIGQUIT);
+		kore_log(LOG_ERR, "=====================================");
+		return;
+	}
 
 	/* Set worker_max_connections for kore_connection_init(). */
 	tmp = worker_max_connections;
@@ -901,35 +878,47 @@ kore_server_start(int argc, char *argv[])
 	kore_platform_event_init();
 	kore_msg_parent_init();
 
-	quit = 0;
 	worker_max_connections = tmp;
 
 	kore_timer_init();
+
 #if !defined(KORE_NO_HTTP)
-	kore_timer_add(kore_accesslog_run, 100, NULL, 0);
+	alog = 0;
+
+	LIST_FOREACH(srv, &kore_servers, list) {
+		TAILQ_FOREACH(dom, &srv->domains, list) {
+			if (dom->accesslog != -1)
+				alog = 1;
+		}
+	}
+
+	if (alog) {
+		kore_timer_add(kore_accesslog_run, 100, NULL, 0);
+		kore_log(LOG_INFO, "accesslog vacuum is enabled");
+	}
 #endif
 
 #if defined(KORE_USE_PYTHON)
 	kore_msg_unregister(KORE_PYTHON_SEND_OBJ);
 #endif
 
-	while (quit != 1) {
-		if (sig_recv != 0) {
-			last_sig = sig_recv;
+	while (kore_quit == KORE_QUIT_NONE) {
+		last_sig = sig_recv;
 
-			switch (sig_recv) {
+		if (last_sig != 0) {
+			switch (last_sig) {
 			case SIGHUP:
-				kore_worker_dispatch_signal(sig_recv);
+				kore_worker_dispatch_signal(last_sig);
 				kore_module_reload(0);
 				break;
 			case SIGINT:
 			case SIGQUIT:
 			case SIGTERM:
-				quit = 1;
-				kore_worker_dispatch_signal(sig_recv);
+				kore_quit = KORE_QUIT_NORMAL;
+				kore_worker_dispatch_signal(last_sig);
 				continue;
 			case SIGUSR1:
-				kore_worker_dispatch_signal(sig_recv);
+				kore_worker_dispatch_signal(last_sig);
 				break;
 			case SIGCHLD:
 				kore_worker_reap();
@@ -948,8 +937,20 @@ kore_server_start(int argc, char *argv[])
 		kore_platform_event_wait(netwait);
 		kore_connection_prune(KORE_CONNECTION_PRUNE_DISCONNECT);
 		kore_timer_run(kore_time_ms());
+		kore_worker_reap();
 	}
 
+	kore_worker_dispatch_signal(SIGQUIT);
+}
+
+static void
+kore_server_shutdown(void)
+{
+	if (!kore_quiet)
+		kore_log(LOG_INFO, "server shutting down");
+
+	kore_worker_shutdown();
+
 #if !defined(KORE_NO_HTTP)
 	kore_accesslog_gather(NULL, kore_time_ms(), 1);
 #endif
@@ -957,6 +958,7 @@ kore_server_start(int argc, char *argv[])
 	kore_platform_event_cleanup();
 	kore_connection_cleanup();
 	kore_domain_cleanup();
+	kore_tls_cleanup();
 	net_cleanup();
 }
 
